@@ -57,6 +57,8 @@ class Cloner(object):
         self._clone_uuid = None
         self._clone_sparse = True
         self._clone_xml = None
+        self.clone_nvram = None
+        self._nvram_disk = None
 
         self._force_target = []
         self._skip_target = []
@@ -103,7 +105,7 @@ class Cloner(object):
             Guest.validate_name(self.conn, name,
                                 check_collision=not self.replace,
                                 validate=False)
-        except ValueError, e:
+        except ValueError as e:
             raise ValueError(_("Invalid name for new guest: %s") % e)
 
         self._clone_name = name
@@ -113,7 +115,7 @@ class Cloner(object):
     def set_clone_uuid(self, uuid):
         try:
             util.validate_uuid(uuid)
-        except ValueError, e:
+        except ValueError as e:
             raise ValueError(_("Invalid uuid for new guest: %s") % e)
 
         if util.vm_uuid_collision(self.conn, uuid):
@@ -145,7 +147,7 @@ class Cloner(object):
                     disk.set_vol_install(vol_install)
                 disk.validate()
                 disklist.append(disk)
-            except Exception, e:
+            except Exception as e:
                 logging.debug("Error setting clone path.", exc_info=True)
                 raise ValueError(_("Could not use path '%s' for cloning: %s") %
                                  (path, str(e)))
@@ -340,7 +342,7 @@ class Cloner(object):
             # replace vol_install with a CloneVolume instance, otherwise
             # simply set input_vol on the dest vol_install
             if (clone_vol_install.pool.name() ==
-                orig_disk.get_vol_object().storagePoolLookupByVolume().name()):
+                orig_disk.get_parent_pool().name()):
                 vol_install = StorageVolume(self.conn)
                 vol_install.input_vol = orig_disk.get_vol_object()
                 vol_install.sync_input_vol()
@@ -361,6 +363,36 @@ class Cloner(object):
         clone_disk.validate()
 
 
+    def _prepare_nvram(self):
+        if self.clone_nvram is None:
+            nvram_dir = os.path.dirname(self._guest.os.nvram)
+            self.clone_nvram = os.path.join(nvram_dir,
+                                            "%s_VARS.fd" % self._clone_name)
+
+        nvram = VirtualDisk(self.conn)
+        nvram.path = self.clone_nvram
+        if (not self.preserve_dest_disks and
+            nvram.wants_storage_creation()):
+
+            old_nvram = VirtualDisk(self.conn)
+            old_nvram.path = self._guest.os.nvram
+            if not old_nvram.get_vol_object():
+                raise RuntimeError(_("Path does not exist: %s") %
+                                     old_nvram.path)
+
+            nvram_install = VirtualDisk.build_vol_install(
+                    self.conn, os.path.basename(nvram.path),
+                    nvram.get_parent_pool(), nvram.get_size(), False)
+            nvram_install.input_vol = old_nvram.get_vol_object()
+            nvram_install.sync_input_vol(only_format=True)
+            nvram_install.reflink = self.reflink
+            nvram.set_vol_install(nvram_install)
+
+        nvram.validate()
+        self._nvram_disk = nvram
+        self._guest.os.nvram = nvram.path
+
+
     def setup_clone(self):
         """
         Validate and set up all parameters needed for the new (clone) VM
@@ -382,7 +414,7 @@ class Cloner(object):
         self._clone_macs.reverse()
         for dev in self._guest.get_devices("graphics"):
             if dev.port and dev.port != -1:
-                logging.warn(_("Setting the graphics device port to autoport, "
+                logging.warning(_("Setting the graphics device port to autoport, "
                                "in order to avoid conflicting."))
                 dev.port = -1
 
@@ -420,18 +452,13 @@ class Cloner(object):
             if channel.type == VirtualChannelDevice.TYPE_UNIX:
                 channel.source_path = None
 
+        if self._guest.os.nvram:
+            self._prepare_nvram()
+
         # Save altered clone xml
         self._clone_xml = self._guest.get_xml_config()
         logging.debug("Clone guest xml is\n%s", self._clone_xml)
 
-    def setup(self):
-        """
-        Helper function that wraps setup_original and setup_clone, with
-        additional debug logging.
-        """
-        self.setup_original()
-        self.setup_clone()
-
     def start_duplicate(self, meter=None):
         """
         Actually perform the duplication: cloning disks if needed and defining
@@ -452,7 +479,9 @@ class Cloner(object):
             if self.preserve:
                 for dst_dev in self.clone_disks:
                     dst_dev.setup(meter=meter)
-        except Exception, e:
+                if self._nvram_disk:
+                    self._nvram_disk.setup(meter=meter)
+        except Exception as e:
             logging.debug("Duplicate failed: %s", str(e))
             if dom:
                 dom.undefine()
@@ -546,7 +575,7 @@ class Cloner(object):
                     if newd.wants_storage_creation():
                         raise ValueError(_("Disk path '%s' does not exist.") %
                                          newd.path)
-            except Exception, e:
+            except Exception as e:
                 logging.debug("Exception creating clone disk objects",
                     exc_info=True)
                 raise ValueError(_("Could not determine original disk "
