@@ -32,7 +32,8 @@ class VirtualFilesystem(VirtualDevice):
     TYPE_BLOCK = "block"
     TYPE_RAM = "ram"
     TYPE_DEFAULT = "default"
-    TYPES = [TYPE_MOUNT, TYPE_TEMPLATE, TYPE_FILE, TYPE_BLOCK, TYPE_RAM, TYPE_DEFAULT]
+    TYPES = [TYPE_MOUNT, TYPE_TEMPLATE, TYPE_FILE, TYPE_BLOCK, TYPE_RAM,
+        TYPE_DEFAULT]
 
     MODE_PASSTHROUGH = "passthrough"
     MODE_MAPPED = "mapped"
@@ -49,13 +50,14 @@ class VirtualFilesystem(VirtualDevice):
     DRIVER_LOOP = "loop"
     DRIVER_NBD = "nbd"
     DRIVER_DEFAULT = "default"
-    DRIVERS = [DRIVER_PATH, DRIVER_HANDLE, DRIVER_LOOP, DRIVER_NBD, DRIVER_DEFAULT]
+    DRIVERS = [DRIVER_PATH, DRIVER_HANDLE, DRIVER_LOOP, DRIVER_NBD,
+        DRIVER_DEFAULT]
 
 
-    type = XMLProperty("./@type",
+    _type_prop = XMLProperty("./@type",
                        default_cb=lambda s: None,
                        default_name=TYPE_DEFAULT)
-    mode = XMLProperty("./@accessmode",
+    accessmode = XMLProperty("./@accessmode",
                        default_cb=lambda s: None,
                        default_name=MODE_DEFAULT)
     wrpolicy = XMLProperty("./driver/@wrpolicy",
@@ -74,7 +76,7 @@ class VirtualFilesystem(VirtualDevice):
         # In case of qemu for default fs type (mount) target is not
         # actually a directory, it is merely a arbitrary string tag
         # that is exported to the guest as a hint for where to mount
-        if (self.conn.is_qemu() and
+        if ((self.conn.is_qemu() or self.conn.is_test()) and
             (self.type is None or
              self.type == self.TYPE_DEFAULT or
              self.type == self.TYPE_MOUNT)):
@@ -109,5 +111,39 @@ class VirtualFilesystem(VirtualDevice):
         return setattr(self, self._type_to_source_prop(), val)
     source = property(_get_source, _set_source)
 
+    def _get_type(self):
+        return getattr(self, '_type_prop')
+    def _set_type(self, val):
+        # Get type/value of the attrubute of "source" property
+        old_source_type = self._type_to_source_prop()
+        old_source_value = self.source
+
+        # Update "type" property
+        new_type = setattr(self, '_type_prop', val)
+
+        # If the attribute type of 'source' property has changed
+        # restore the value
+        if old_source_type != self._type_to_source_prop():
+            self.source = old_source_value
+
+        return new_type
+
+    type = property(_get_type, _set_type)
+
+    def set_defaults(self, guest):
+        ignore = guest
+
+        if self.conn.is_qemu() or self.conn.is_lxc() or self.conn.is_test():
+            # type=mount is the libvirt default. But hardcode it
+            # here since we need it for the accessmode check
+            if self.type is None or self.type == self.TYPE_DEFAULT:
+                self.type = self.TYPE_MOUNT
+
+            # libvirt qemu defaults to accessmode=passthrough, but that
+            # really only works well for qemu running as root, which is
+            # not the common case. so use mode=mapped
+            if self.accessmode is None or self.accessmode == self.MODE_DEFAULT:
+                self.accessmode = self.MODE_MAPPED
+
 
 VirtualFilesystem.register_type()
